package com.ijchart.xychart.shape ;

import java.awt.Graphics2D ;
import java.awt.Shape ;
import java.awt.geom.AffineTransform ;
import java.awt.geom.Arc2D ;
import java.awt.geom.Area ;
import java.awt.geom.GeneralPath ;
import java.awt.geom.PathIterator ;
import java.awt.geom.Point2D ;
import java.awt.geom.Rectangle2D ;
import java.io.Serializable ;
import java.util.ArrayList ;
import java.util.List ;

import com.ijchart.xychart.constant.ShapeType ;
import com.ijchart.xychart.ui.Graphics2DState ;
import com.ijchart.xychart.util.GeometryUtil ;
import com.ijchart.xychart.util.StringUtil ;
import com.ijchart.xychart.util.StrokeUtil ;
import com.ijchart.xychart.util.TextUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ͼ.</zh>
 * <en>Draw a sector shape</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class RingChartShape extends SectorChartShape implements Cloneable,
  Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -5714745448985058124L ;

  /**
   * <zh>ȱʡĻĳ뾶ռ.</zh>
   * <en>The default length factor (ring-width/sector-radius-width).</en>
   */
  public static final double DEFAULT_LENGTH_FACTOR = 0.3 ;

  /**
   * <zh>ĳ뾶ռ.</zh>
   * <en>The length factor (ring-width/sector-radius-width).</en>
   */
  private double lengthFactor ;

  /**
   * <zh>.</zh>
   * <en>A constructor.</en>
   *
   * @param arc Arc2D <zh>ͼ (Ϊnull).</zh>
   * <en>A arc shape (null not permitted).</en>
   * @param drawArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The draw area (null not permitted).</en>
   * @param text String <zh>Ҫʾ (Ϊnull).</zh>
   * <en>The text (null permitted).</en>
   */
  public RingChartShape(Arc2D arc, Rectangle2D drawArea, String text) {
    super(arc, drawArea, text, false) ;
    this.lengthFactor = DEFAULT_LENGTH_FACTOR ;
  }

  /**
   * <zh>ػĳ뾶ռ.</zh>
   * <en>Returns the length factor.</en>
   *
   * @return double <zh>.</zh>
   * <en>The length factor.</en>
   */
  public double getLengthFactor() {
    return this.lengthFactor ;
  }

  /**
   * <zh>ûĳ뾶ռ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the length factor and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param factor double <zh>,(0,1].</zh>
   * <en>The length factor,in interval (0,1].</en>
   */
  public void setLengthFactor(double factor) {
    if (factor <= 0 || factor > 1) {
      throw new IllegalArgumentException("Argument 'factor' must be in interval (0,1].") ;
    }

    if (this.lengthFactor != factor) {
      this.lengthFactor = factor ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ͼ豸(Ļӡ)ϻƴͼ.ʱshapeλ.</zh>
   * <en>Draw the ChartShape on a Java 2d graphics device(such as the screen or the printer).
   * The shape's coordinate is ignored.</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(Ļӡ) (Ϊnull)</zh>
   * <en>The graphics device (such as the screen or the printer) (null not permitted).</en>
   */
  public void draw(Graphics2D g2) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }
    if (!isVisible()) {
      return ;
    }

    Graphics2DState state = new Graphics2DState(g2) ;
    state.saveOldState() ;

    Arc2D arc = getArc() ;
    double angleStart = arc.getAngleStart() ; //ʼǶ
    double angleExtent = arc.getAngleExtent() ;
    double arcX = arc.getX() ;
    double arcY = arc.getY() ;
    double arcWidth = arc.getWidth() ;
    double arcHeight = arc.getHeight() ;
    Arc2D newArc = new Arc2D.Double(0.0D, 0.0D, arcWidth, arcHeight,
                                    angleStart, angleExtent, arc.getArcType()) ;
    double arcHoleLen = 1 - this.lengthFactor ;
    double halfFactor = this.lengthFactor * 0.5 ;
    Arc2D arcHole = new Arc2D.Double(arcWidth * halfFactor, arcHeight * halfFactor,
                                     arcWidth * arcHoleLen, arcHeight * arcHoleLen, angleStart,
                                     360, arc.getArcType()) ; //360֤ȥԲ

    g2.translate(arcX, arcY) ;
    g2.rotate(getRotate()) ;
    beforeDraw(g2) ;
//    g2.setClip(newArc) ;

    Area ringArea = new Area(newArc) ;
    ringArea.subtract(new Area(arcHole)) ;

    //Ӱ
    if (getSectorShadowPaint() != null) {
      AffineTransform transform = AffineTransform.getTranslateInstance(getSectorShadowXOffset(),
                                                                       getSectorShadowYOffset()) ;
      Shape shadowRingArea = transform.createTransformedShape(ringArea) ;
      g2.setPaint(getSectorShadowPaint()) ;
      g2.fill(shadowRingArea) ;
    }

    drawBackground(g2, ringArea) ;
    drawBorder(g2, ringArea) ;

    //Ʊǩ
    String label = getText() ;
    if (isDrawText() && !StringUtil.isEmpty(label)) {
      g2.setFont(getTextFont()) ;
      Rectangle2D bounds = TextUtil.getStringBoundWithFm(g2, label) ;
      Rectangle2D out = getLabelPadding().createOutsetRectangle(bounds) ;

      Arc2D middleArc = new Arc2D.Double(arc.getX(), arc.getY(), arc.getWidth(), arc.getHeight(),
                                         arc.getAngleStart(), arc.getAngleExtent() / 2, Arc2D.PIE) ;
      Point2D arcMiddle = middleArc.getEndPoint() ;
      Point2D arcCenter = new Point2D.Double(arc.getCenterX(), arc.getCenterY()) ;
      double dis = GeometryUtil.distanceOfPoints(arcCenter, arcMiddle) * (1 - this.lengthFactor / 2) ;
      Point2D labelPosition = GeometryUtil.extentPoint(arcCenter, arcMiddle, dis) ;

      float strokeWidth = StrokeUtil.getStrokeWidth(getLabelOutlineStroke()) ;
      Rectangle2D labelBound = new Rectangle2D.Double(labelPosition.getX() - arcX
                                                      - 0.5 * out.getWidth(),
                                                      labelPosition.getY() - arcY
                                                      - 0.5 * out.getHeight(),
                                                      out.getWidth() + 2 * strokeWidth,
                                                      out.getHeight() + 2 * strokeWidth) ;

      //draw label shadow
      if (getLabelShadowPaint() != null) {
        Rectangle2D shadowBound = new Rectangle2D.Double(labelBound.getX() + getLabelShadowXOffset(),
                                                         labelBound.getY() + getLabelShadowYOffset(),
                                                         labelBound.getWidth(),
                                                         labelBound.getHeight()) ;
        g2.setPaint(getLabelShadowPaint()) ;
        g2.fill(shadowBound) ;
      }

      //draw label background
      if (getLabelBackgroundPaint() != null) {
        g2.setPaint(getLabelBackgroundPaint()) ;
        g2.fill(labelBound) ;
      }

      //draw label border
      if (getLabelOutlinePaint() != null && getLabelOutlineStroke() != null) {
        g2.setPaint(getLabelOutlinePaint()) ;
        g2.setStroke(getLabelOutlineStroke()) ;
        g2.draw(labelBound) ;
      }

      drawTextWithTrimMarginRect(g2, labelBound) ;
    }

    afterDraw(g2) ;
    g2.rotate( -getRotate()) ;
    state.restoreOldState() ;
    drawSelect(g2) ;
  }

  /**
   * <zh>ʱͼεĵ.</zh>
   * <en>Returns the lock coordinates of this shape.</en>
   *
   * @return Point2D <zh>ֵ.</zh>
   * <en>The coordinate.</en>
   */
  public Point2D getLockPoint() {
    Arc2D arc = getArc() ;
    if (arc == null) {
      return null ;
    }

    double arcX = arc.getX() ;
    double arcY = arc.getY() ;

    Arc2D middleArc = new Arc2D.Double(arc.getX(), arc.getY(), arc.getWidth(), arc.getHeight(),
                                       arc.getAngleStart(), arc.getAngleExtent() / 2, Arc2D.PIE) ;
    Point2D arcMiddle = middleArc.getEndPoint() ;
    Point2D arcCenter = new Point2D.Double(arc.getCenterX(), arc.getCenterY()) ;
    double dis = GeometryUtil.distanceOfPoints(arcCenter, arcMiddle) * (1 - this.lengthFactor / 2) ;
    Point2D labelPos = GeometryUtil.extentPoint(arcCenter, arcMiddle, dis) ;

    Rectangle2D labelBound = new Rectangle2D.Double(labelPos.getX() - arcX, labelPos.getY() - arcY,
                                                    1, 1) ;

    AffineTransform transform = AffineTransform.getTranslateInstance(arcX, arcY) ;
    transform.rotate(getRotate()) ;
    Shape rotateShape = transform.createTransformedShape(labelBound) ;
    Rectangle2D rotateBound = rotateShape.getBounds2D() ;

    return new Point2D.Double(rotateBound.getCenterX(), rotateBound.getCenterY()) ;
  }

  /**
   * <zh>任ͼ.</zh>
   * <en>Calculate the transformed shape.</en>
   */
  protected void calculateTransformedShape() {
    if (transformedShape == null) {
      Arc2D arc = getArc() ;
      double angleStart = arc.getAngleStart() ; //ʼǶ
      double angleExtent = arc.getAngleExtent() ;
      double arcX = arc.getX() ;
      double arcY = arc.getY() ;
      double arcWidth = arc.getWidth() ;
      double arcHeight = arc.getHeight() ;
      Arc2D newArc = new Arc2D.Double(0.0D, 0.0D, arcWidth, arcHeight,
                                      angleStart, angleExtent, arc.getArcType()) ;

      double halfFactor = this.lengthFactor * 0.5 ;
      double arcHoleLen = 1 - this.lengthFactor ;
      Arc2D arcHole = new Arc2D.Double(arcWidth * halfFactor, arcHeight * halfFactor,
                                       arcWidth * arcHoleLen, arcHeight * arcHoleLen, angleStart,
                                       360, arc.getArcType()) ; //5Ǳ֤㹻

      Area ringArea = new Area(newArc) ;
      ringArea.subtract(new Area(arcHole)) ;

      AffineTransform transform = new AffineTransform() ;
      transform.translate(arcX, arcY) ;
      transform.rotate(getRotate()) ;
      transformedShape = transform.createTransformedShape(ringArea) ;
    }
  }

  /**
   * <zh>ȵͼ.</zh>
   * <en>Returns the hotspot shape.</en>
   *
   * @return Shape <zh>ȵͼ (Ϊnull).</zh>
   * <en>The hotspot shape (maybe null).</en>
   */
  protected Shape getHotspotShape() {
    Arc2D arc = getArc() ;
    double angleStart = arc.getAngleStart() ; //ʼǶ
    double angleExtent = Math.min(arc.getAngleExtent(), 360) ;
    double arcWidth = arc.getWidth() ;
    double arcHeight = arc.getHeight() ;
    double arcHoleLen = 1 - this.lengthFactor ;
    double halfFactor = this.lengthFactor * 0.5 ;
    Arc2D arcHole = new Arc2D.Double(0, 0, arcWidth * arcHoleLen, arcHeight * arcHoleLen,
                                     angleStart, angleExtent, arc.getArcType()) ;

    GeneralPath outPath = GeometryUtil.getArcPath(arc) ;
    GeneralPath inPath = GeometryUtil.getArcPath(arcHole) ;

    List inPoints = new ArrayList() ;
    PathIterator inPathIter = inPath.getPathIterator(null) ;
    float[] coords = new float[6] ;
    while (!inPathIter.isDone()) {
      inPathIter.currentSegment(coords) ;
      inPoints.add(new float[] {coords[0], coords[1]}) ;
      inPathIter.next() ;
    }

    float arcHoleX = (float) (arcWidth * halfFactor) ;
    float arcHoleY = (float) (arcHeight * halfFactor) ;
    for (int i = inPoints.size() - 1 ; i >= 0 ; i--) {
      float[] point = (float[]) inPoints.get(i) ;
      outPath.lineTo(point[0] + arcHoleX, point[1] + arcHoleY) ;
    }
    outPath.closePath() ;

    AffineTransform transform = AffineTransform.getTranslateInstance(arc.getX(), arc.getY()) ;
    transform.rotate(getRotate()) ;

    return transform.createTransformedShape(outPath) ;
  }

  /**
   * <zh>ͼε.</zh>
   * <en>The type for describe this shape.</en>
   * @return ShapeType <zh>{@link com.ijchart.xychart.constant.ShapeType}ඨĳ
   * (Ϊnull).</zh>
   * <en>The constant value in class {@link com.ijchart.xychart.constant.ShapeType}.</en>
   */
  public ShapeType getShapeType() {
    return ShapeType.SHAPE_RING_SECTOR ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double result = super.hashCode() ;
    result = 37 * result + 100 * this.lengthFactor ;

    return (int) result ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Indicates whether some other object is "equal to" this one.</en>
   *
   * @param obj <zh>ҪȽϵĶ</zh>
   * <en>the reference object with which to compare.</en>
   * @return <zh>ȷtrue,򷵻false.</zh>
   * <en>true if this object is the same as the obj argument; false otherwise.</en>
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true ;
    }
    if (!(obj instanceof RingChartShape)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    RingChartShape that = (RingChartShape) obj ;
    if (this.lengthFactor != that.lengthFactor) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

}
